@@ -1,15 +1,16 @@
 package Apache::DBI;
 use strict;
 
-# $Id: DBI.pm,v 1.5 2003/01/11 03:17:49 ask Exp $
+# $Id: DBI.pm,v 1.12 2004/02/18 00:18:50 ask Exp $
 
 BEGIN { eval { require Apache } }
+BEGIN { eval { require mod_perl2; require Apache2::Module; } }
 use DBI ();
 use Carp qw(carp);
 
 require_version DBI 1.00;
 
-$Apache::DBI::VERSION = '0.90_02';
+$Apache::DBI::VERSION = '0.100';
 
 # 1: report about new connect
 # 2: full debug output
@@ -21,8 +22,9 @@ my @ChildConnect; # connections to be established when a new httpd child is crea
 my %Rollback;     # keeps track of pushed PerlCleanupHandler which can do a rollback after the request has finished
 my %PingTimeOut;  # stores the timeout values per data_source, a negative value de-activates ping, default = 0
 my %LastPingTime; # keeps track of last ping per data_source
-my $Idx;          # key of %Connected and %Rollback.
 
+# Check to see if we need to reset TaintIn and TaintOut
+my $TaintInOut = ($DBI::VERSION>=1.31)?1:0;
 
 # supposed to be called in a startup script.
 # stores the data_source of all connections, which are supposed to be created upon
@@ -31,11 +33,19 @@ my $Idx;          # key of %Connected and %Rollback.
 sub connect_on_init { 
     # provide a handler which creates all connections during server startup
 
-    # TODO - Should check for mod_perl 2 and do the right thing there
-    carp "Apache.pm was not loaded\n" and return unless $INC{'Apache.pm'};
-    if(!@ChildConnect and Apache->can('push_handlers')) {
-        Apache->push_handlers(PerlChildInitHandler => \&childinit);
-    }
+		if ($ENV{MOD_PERL_API_VERSION} == 2) {
+			if (!@ChildConnect) {
+				require Apache2::ServerUtil;
+				my $s = Apache2::ServerUtil->server;
+				$s->push_handlers(PerlChildInitHandler => \&childinit);
+			}
+		}
+		else {
+   		carp "Apache.pm was not loaded\n" and return unless $INC{'Apache.pm'};
+   		if (!@ChildConnect and Apache->can('push_handlers')) {
+     	  Apache->push_handlers(PerlChildInitHandler => \&childinit);
+   		}
+		}
     # store connections
     push @ChildConnect, [@_];
 }
@@ -67,12 +77,17 @@ sub connect {
     my $dsn    = "dbi:$drh->{Name}:$args[0]";
     my $prefix = "$$ Apache::DBI            ";
 
-    $Idx = join $;, $args[0], $args[1], $args[2];
+    my $Idx = join $;, $args[0], $args[1], $args[2]; # key of %Connected and %Rollback.
 
     # the hash-reference differs between calls even in the same
     # process, so de-reference the hash-reference 
     if (3 == $#args and ref $args[3] eq "HASH") {
-      map { $Idx .= "$;$_=$args[3]->{$_}" } sort keys %{$args[3]};
+      # should we default to '__undef__' or something for undef values?
+      map { $Idx .= "$;$_=" . 
+	      (defined $args[3]->{$_} 
+	       ? $args[3]->{$_} 
+	       : '')
+	    } sort keys %{$args[3]};
     } elsif (3 == $#args) {
         pop @args;
     }
@@ -81,18 +96,36 @@ sub connect {
     # won't be useful after ChildInit, since multiple processes trying to
     # work over the same database connection simultaneously will receive
     # unpredictable query results.
-    if ($Apache::ServerStarting == 1) {
-        print STDERR "$prefix skipping connection during server startup, read the docu !!\n" if $Apache::DBI::DEBUG > 1;
-        return $drh->connect(@args);
+
+    ## See: http://perl.apache.org/docs/2.0/user/porting/compat.html#C__Apache__Server__Starting__and_C__Apache__Server__ReStarting_
+    if (exists $ENV{MOD_PERL_API_VERSION} && $ENV{MOD_PERL_API_VERSION} == 2) {
+        require Apache2::ServerUtil;
+        if (Apache2::ServerUtil::restart_count() == 1) {
+          print STDERR "$prefix skipping connection during server startup, read the docu !!\n" if $Apache::DBI::DEBUG > 1;
+        } 
+    }
+    else {
+        if ($Apache::ServerStarting and $Apache::ServerStarting == 1) {
+            print STDERR "$prefix skipping connection during server startup, read the docu !!\n" if $Apache::DBI::DEBUG > 1;
+            return $drh->connect(@args);
+        }   
     }
 
-    # this PerlCleanupHandler is supposed to initiate a rollback after the script has finished if AutoCommit is off.
-    my $needCleanup = ($Idx =~ /AutoCommit[^\d]+0/) ? 1 : 0;
-    # TODO - Fix mod_perl 2.0 here
-    if(!$Rollback{$Idx} and $needCleanup and Apache->can('push_handlers')) {
+    # this PerlCleanupHandler is supposed to initiate a rollback after the
+    # script has finished if AutoCommit is off.  however, cleanup can only
+    # be determined at end of handle life as begin_work may have been called
+    # to temporarily turn off AutoCommit.
+    if(!$Rollback{$Idx} and Apache->can('push_handlers')) {
         print STDERR "$prefix push PerlCleanupHandler \n" if $Apache::DBI::DEBUG > 1;
-        Apache->push_handlers("PerlCleanupHandler", \&cleanup);
-        # make sure, that the rollback is called only once for every 
+        if ($ENV{MOD_PERL_API_VERSION} == 2) {
+            require Apache2::ServerUtil;
+            my $s = Apache2::ServerUtil->server;
+            $s->push_handlers("PerlCleanupHandler", sub { cleanup($Idx) });
+        }
+        else {
+            Apache->push_handlers("PerlCleanupHandler", sub { cleanup($Idx) });
+        }
+        # make sure, that the rollback is called only once for every
         # request, even if the script calls connect more than once
         $Rollback{$Idx} = 1;
     }
@@ -101,7 +134,11 @@ sub connect {
     $PingTimeOut{$dsn}  = 0 unless $PingTimeOut{$dsn};
     $LastPingTime{$dsn} = 0 unless $LastPingTime{$dsn};
     my $now = time;
-    my $needping = (($PingTimeOut{$dsn} == 0 or $PingTimeOut{$dsn} > 0) and $now - $LastPingTime{$dsn} > $PingTimeOut{$dsn}) ? 1 : 0;
+    # Must ping if TimeOut = 0 else base on time
+    my $needping = ($PingTimeOut{$dsn} == 0 or
+                    ($PingTimeOut{$dsn} > 0 and
+                     $now - $LastPingTime{$dsn} > $PingTimeOut{$dsn})
+                   ) ? 1 : 0;
     print STDERR "$prefix need ping: ", $needping == 1 ? "yes" : "no", "\n" if $Apache::DBI::DEBUG > 1;
     $LastPingTime{$dsn} = $now;
 
@@ -112,6 +149,10 @@ sub connect {
     # RaiseError being on and the handle is invalid.
     if ($Connected{$Idx} and (!$needping or eval{$Connected{$Idx}->ping})) {
         print STDERR "$prefix already connected to '$Idx'\n" if $Apache::DBI::DEBUG > 1;
+
+        # Force clean up of handle in case previous transaction failed to clean up the handle
+        &reset_startup_state($Idx);
+
         return (bless $Connected{$Idx}, 'Apache::DBI::db');
     }
 
@@ -121,6 +162,9 @@ sub connect {
     $Connected{$Idx} = $drh->connect(@args);
     return undef if !$Connected{$Idx};
 
+    # store the parameters of the initial connection in the handle
+    set_startup_state($Idx);
+
     # return the new database handle
     print STDERR "$prefix new connect to '$Idx'\n" if $Apache::DBI::DEBUG;
     return (bless $Connected{$Idx}, 'Apache::DBI::db');
@@ -148,6 +192,7 @@ sub childinit {
 # Note: the PerlCleanupHandler runs after the response has been sent to the client
 
 sub cleanup {
+    my $Idx = shift;
     my $prefix = "$$ Apache::DBI            ";
     print STDERR "$prefix PerlCleanupHandler \n" if $Apache::DBI::DEBUG > 1;
     my $dbh = $Connected{$Idx};
@@ -158,6 +203,49 @@ sub cleanup {
     1;
 }
 
+# Store the default start state of each dbh in the handle
+# Note: This uses private_Apache_DBI hash ref to store it in the handle itself
+
+sub set_startup_state {
+    my $Idx = shift;
+    foreach my $key qw{ AutoCommit Warn CompatMode InactiveDestroy 
+                        PrintError RaiseError HandleError 
+                        ShowErrorStatement TraceLevel FetchHashKeyName 
+                        ChopBlanks LongReadLen LongTruncOk 
+                        Taint Profile} {
+        $Connected{$Idx}->{private_Apache_DBI}{$key} = $Connected{$Idx}->{$key};
+    }
+    if ($TaintInOut) {
+        foreach my $key qw{ TaintIn TaintOut } {
+            $Connected{$Idx}->{private_Apache_DBI}{$key} = $Connected{$Idx}->{$key};
+        }
+    }
+    1;
+}
+
+
+# Restore the default start state of each dbh
+
+sub reset_startup_state {
+    my $Idx = shift;
+    # Rollback current transaction if currently in one
+    $Connected{$Idx}->{Active} and !$Connected{$Idx}->{AutoCommit} and eval {$Connected{$Idx}->rollback};
+
+    foreach my $key qw{ AutoCommit Warn CompatMode InactiveDestroy 
+                        PrintError RaiseError HandleError 
+                        ShowErrorStatement TraceLevel FetchHashKeyName 
+                        ChopBlanks LongReadLen LongTruncOk 
+                        Taint Profile } {
+        $Connected{$Idx}->{$key} = $Connected{$Idx}->{private_Apache_DBI}{$key};
+    }
+    if ($TaintInOut) {
+        foreach my $key qw{ TaintIn TaintOut } {
+            $Connected{$Idx}->{$key} = $Connected{$Idx}->{private_Apache_DBI}{$key};
+        }
+    }
+    1;
+}
+
 
 # This function can be called from other handlers to perform tasks on all cached database handles.
 
@@ -186,9 +274,8 @@ sub all_handlers {
 
 
 # prepare menu item for Apache::Status
-
-Apache::Status->menu_item(
-
+if (exists $ENV{MOD_PERL_API_VERSION} && $ENV{MOD_PERL_API_VERSION} == 2) {
+	Apache2::Status->menu_item(
     'DBI' => 'DBI connections',
     sub {
         my($r, $q) = @_;
@@ -199,11 +286,25 @@ Apache::Status->menu_item(
         push @s, '</TABLE>';
         return \@s;
    }
+	) if Apache2::Module::loaded('Apache2::Status');
+}
+else {
+	Apache::Status->menu_item(
+		'DBI' => 'DBI connections',
+		sub {
+        my($r, $q) = @_;
+        my(@s) = qw(<TABLE><TR><TD>Datasource</TD><TD>Username</TD></TR>);
+        for (keys %Connected) {
+            push @s, '<TR><TD>', join('</TD><TD>', (split($;, $_))[0,1]), "</TD></TR>\n";
+        }
+        push @s, '</TABLE>';
+        return \@s;
+   }
 
-) if ($INC{'Apache.pm'} and Apache->can('module') and Apache->module('Apache::Status'));
- 
-
-
+	) if ($INC{'Apache.pm'}                      # is Apache.pm loaded?
+      and Apache->can('module')               # really?
+      and Apache->module('Apache::Status'));  # Apache::Status too?
+}
 
 1;
 
@@ -273,21 +374,20 @@ which is cheap and safe or you can deactivate the usage of the ping method
 
 Here is generalized ping method, which can be added to the driver module:
 
-{   package DBD::xxx::db; # ====== DATABASE ======
-    use strict;
+   package DBD::xxx::db; # ====== DATABASE ======
+   use strict;
 
-    sub ping {
-        my($dbh) = @_;
-        my $ret = 0;
-        eval {
-            local $SIG{__DIE__}  = sub { return (0); };
-            local $SIG{__WARN__} = sub { return (0); };
-            # adapt the select statement to your database:
-            $ret = $dbh->do('select 1');
-        };
-        return ($@) ? 0 : $ret;
-    }
-}
+   sub ping {
+     my ($dbh) = @_;
+     my $ret = 0;
+     eval {
+       local $SIG{__DIE__}  = sub { return (0); };
+       local $SIG{__WARN__} = sub { return (0); };
+       # adapt the select statement to your database:
+       $ret = $dbh->do('select 1');
+     };
+     return ($@) ? 0 : $ret;
+   }
 
 Transactions: a standard DBI script will automatically perform a rollback
 whenever the script exits. In the case of persistent database connections,
@@ -301,16 +401,18 @@ every script. In order to avoid inconsistencies in the database in case
 AutoCommit is off and the script finishes without an explicit rollback, the 
 Apache::DBI module uses a PerlCleanupHandler to issue a rollback at the
 end of every request. Note, that this CleanupHandler will only be used, if 
-the initial data_source sets AutoCommit = 0. It will not be used, if AutoCommit 
-will be turned off, after the connect has been done. 
-
-This module plugs in a menu item for Apache::Status. The menu lists the 
-current database connections. It should be considered incomplete because of 
-the limitations explained above. It shows the current database connections 
-for one specific server, the one which happens to serve the current request. 
-Other servers might have other database connections. The Apache::Status module 
-has to be loaded before the Apache::DBI module !
+the initial data_source sets AutoCommit = 0 or AutoCommit is turned off, after 
+the connect has been done (ie begin_work). However, because a connection may have
+set other parameters, the handle is reset to its initial connection state before 
+it is returned for a second time.
 
+This module plugs in a menu item for Apache::Status or Apache2::Status. 
+The menu lists the current database connections. It should be considered 
+incomplete because of the limitations explained above. It shows the current 
+database connections for one specific server, the one which happens to serve the 
+current request.  Other servers might have other database connections. 
+The Apache::Status/Apache2::Status module has to be loaded before the 
+Apache::DBI module !
 
 =head1 CONFIGURATION
 
@@ -322,7 +424,7 @@ Add the following line to your httpd.conf or startup.pl:
 It is important, to load this module before any other modules using DBI ! 
 
 A common usage is to load the module in a startup file via the PerlRequire 
-directive. See eg/startup.pl for an example. 
+directive. See eg/startup.pl and eg/startup2.pl for examples.
 
 There are two configurations which are server-specific and which can be done 
 upon server startup: 
@@ -341,8 +443,8 @@ the validation of the database handle. This can be used for drivers, which
 do not implement the ping-method. Setting the timeout > 0 will ping the 
 database only if the last access was more than timeout seconds before. 
 
-For the menu item 'DBI connections' you need to call Apache::Status BEFORE 
-Apache::DBI ! For an example of the configuration order see startup.pl. 
+For the menu item 'DBI connections' you need to call Apache::Status/Apache2::Status
+BEFORE Apache::DBI ! For an example of the configuration order see startup.pl. 
 
 To enable debugging the variable $Apache::DBI::DEBUG must be set. This 
 can either be done in startup.pl or in the user script. Setting the variable 
@@ -359,8 +461,8 @@ and that mod_perl needs to be configured with the appropriate call-back hooks:
 
 =head1 MOD_PERL 2.0
 
-Apache::DBI version 0.90_02 and later might work under mod_perl 2.0.
-See the Changes file for more information.  Also beware that it has
+Apache::DBI version 0.96 and should work under mod_perl 2.0 RC5 and later.
+See the Changes file for more information.  Beware that it has
 only been tested very lightly.
 
 =head1 SEE ALSO
@@ -371,9 +473,12 @@ L<Apache>, L<mod_perl>, L<DBI>
 =head1 AUTHORS
 
 =item *
-Apache::DBI by Edmund Mergl (now supported and maintained by
-the modperl mailinglist, see the mod_perl documentation for
-instructions on how to subscribe).
+Ask Bjoern Hansen <ask@develooper.com> is currently packaging new releases.
+
+=item *
+Edmund Mergl was the original author of Apache::DBI.  It is now
+supported and maintained by the modperl mailinglist, see the mod_perl
+documentation for instructions on how to subscribe.
 
 =item *
 mod_perl by Doug MacEachern.
